//   DTL -----------------------------------------------------------------------
//   Delphi Template Library
//      Containers module define
//        - STACK
//        - QUEUE
//        - DEQUE
//        - ARRAY
//        - SET
//        - MULTISET
//        - MAP
//        - MULTIMAP

# DTL_DECLARE_LINKEDLIST_ITEM `container `elem   => _BEGIN_
    P\`container\LinkedItem\`elem = ^R\`container\LinkedItem\`elem;
    R\`container\LinkedItem\`elem = record
      value  : `elem;
      next   :  P\`container\LinkedItem\`elem;
      prev   :  P\`container\LinkedItem\`elem;
  _END_;

# DTL_DECLARE_TREENODE_ITEM `container `elem   => _BEGIN_
    P\`container\LinkedItem\`elem = ^R\`container\LinkedItem\`elem;
    R\`container\LinkedItem\`elem = record
      value  : `elem;
      next   :  P\`container\LinkedItem\`elem;
      prev   :  P\`container\LinkedItem\`elem;
      parent :  P\`container\LinkedItem\`elem;
      color  : ENUM_DTL_RB_COLOR;
  _END_;

# DTL_DECLARE_ITERATENEXTPREV_MULTISET `container `elemtype   => _BEGIN_
    private
      function  IterateNextPrev(var iterator : RDtlContainerIterator; IsReversed : boolean) : boolean;
          begin
            with iterator do
                if IsReversed then begin
                    Result := (current_e <> FHeadItem) or (current_i <> 0);
                    if Result then
                      if current_i > 0
                          then current_i := current_i - 1
                          else begin
                                current_e := TreeIteratorPrev(P\`container\LinkedItem\`elemtype(current_e));
                                iterator.current_i := P\`container\LinkedItem\`elemtype(current_e).valuecount - 1;
                               end;
                end else begin
                    Result := (current_e <> FTailItem) or (current_i <> P\`container\LinkedItem\`elemtype(current_e).valuecount - 1);
                    if Result then
                      if current_i <> P\`container\LinkedItem\`elemtype(current_e).valuecount - 1
                          then current_i := current_i + 1
                          else begin
                                current_e := TreeIteratorNext(P\`container\LinkedItem\`elemtype(current_e));
                                iterator.current_i := 0;
                               end;
                end;
          end;
    public
      function  head_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := FHeadItem; isBackward := false; end; end;
      function  tail_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := FTailItem.valuecount - 1; owner := self; current_e := FTailItem; isBackward := true; end; end;
  _END_;

# DTL_DECLARE_ITERATENEXTPREV_TREE `container `elemtype   => _BEGIN_
    private
      function  IterateNextPrev(var iterator : RDtlContainerIterator; IsReversed : boolean) : boolean;
          begin
            with iterator do
                if IsReversed then begin
                    Result := (current_e <> FHeadItem);
                    if Result then current_e := TreeIteratorPrev(P\`container\LinkedItem\`elemtype(current_e));
                end else begin
                    Result := (current_e <> FTailItem);
                    if Result then current_e := TreeIteratorNext(P\`container\LinkedItem\`elemtype(current_e));
                end;
          end;
    public
      function  head_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := FHeadItem; isBackward := false; end; end;
      function  tail_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := FTailItem; isBackward := true; end; end;
  _END_;

# DTL_DECLARE_ITERATENEXTPREV_LINKLIST `container `elemtype   => _BEGIN_
    private
      function  IterateNextPrev(var iterator : RDtlContainerIterator; IsReversed : boolean) : boolean;
          begin
            with iterator do
                if IsReversed then begin
                    Result := (current_e <> FHeadItem);
                    if Result then current_e := P\`container\LinkedItem\`elemtype(iterator.current_e).Prev;
                end else begin
                    Result := (current_e <> FTailItem);
                    if Result then current_e := P\`container\LinkedItem\`elemtype(iterator.current_e).Next;
                end;
          end;
    public
      function  head_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := FHeadItem; isBackward := false; end; end;
      function  tail_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := FTailItem; isBackward := true; end; end;
  _END_;


# DTL_DECLARE_LINKEDLIST_ITERATORS `container `elemtype   => _BEGIN_
    public
      function  next_iterator(var iterator : RDtlContainerIterator; const offset : integer = 1) : boolean;
          begin
            Result := false;
            for `i : integer := 1 to offset do begin
              Result := IterateNextPrev(iterator, iterator.IsBackward);
              if not Result then break;
            end;
          end;
      function  prev_iterator(var iterator : RDtlContainerIterator; const offset : integer = 1) : boolean;
          begin
            Result := false;
            for `i : integer := 1 to offset do begin
              Result := IterateNextPrev(iterator, not iterator.IsBackward);
              if not Result then break;
            end;
          end;
      function  is_iterator_valid(var iterator : RDtlContainerIterator) : boolean;
          begin Result := iterator.current_e <> nil; end

      function  get_value_at(var iterator : RDtlContainerIterator) : `elemtype;
                    begin Result := P\`container\LinkedItem\`elemtype(iterator.current_e).value; end;
      procedure set_value_at(var iterator : RDtlContainerIterator; const value : `elemtype);
                    begin P\`container\LinkedItem\`elemtype(iterator.current_e).value := value; end;

  _END_;

# DTL_DECLARE_INDEXED_ITERATORS `container `elemtype   => _BEGIN_
    protected
      function  is_iterator_in_deep(var iterator : RDtlContainerIterator; const offset : integer = 1) : boolean;
          begin if iterator.IsBackward
                    then Result := iterator.current_i <> 0
                    else Result := iterator.current_i <> FCount - 1; end;
    public
      function  get_value_at(var iterator : RDtlContainerIterator) : `elemtype;
                    begin Result := Items[iterator.current_i]; end;
      procedure set_value_at(var iterator : RDtlContainerIterator; const value : `elemtype);
                    begin Items[iterator.current_i] := value; end;

      function  next_iterator(var iterator : RDtlContainerIterator; const offset : integer = 1) : boolean;
                    begin
                      Result := false;
                      for `i : integer := 1 to offset do begin
                        Result := is_iterator_in_deep(iterator); if not Result then exit;
                        if iterator.IsBackward
                            then iterator.current_i := iterator.current_i - 1
                            else iterator.current_i := iterator.current_i + 1;
                      end;
                    end;
      function  prev_iterator(var iterator : RDtlContainerIterator; const offset : integer = 1) : boolean;
                    begin
                      Result := false;
                      for `i : integer := 1 to offset do begin
                        Result := is_iterator_in_deep(iterator); if not Result then exit;
                        if iterator.IsBackward
                            then iterator.current_i := iterator.current_i + 1
                            else iterator.current_i := iterator.current_i - 1;
                      end;
                    end;
      function  head_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := 0; owner := self; current_e := nil; isBackward := false; end; end;
      function  tail_iterator : RDtlContainerIterator;
                    begin with Result do begin current_i := Count - 1; owner := self; current_e := nil; isBackward := true; end; end;
      function  is_iterator_valid(var iterator : RDtlContainerIterator) : boolean;
          begin Result := (iterator.current_i >= 0) and (iterator.current_i < FCount); end

  _END_;

# DTL_CONTAINER_INNER_ARRAYED_OVERLOAD_FN `container `elemtype `name => _BEGIN_
      function  `name(const values : array of `elemtype) : P\`container\LinkedItem\`elemtype; overload;
                    begin Result := nil; for `i : integer := low(values) to high(values) do Result := `name(values[`i]); end;
  _END_;
# DTL_CONTAINER_INNER_ARRAYED_OVERLOAD_PR `elemtype `name => _BEGIN_
      procedure `name(const values : array of `elemtype); overload;
                    begin for `i : integer := low(values) to high(values) do `name(values[`i]); end;
  _END_;

# DTL_CONTAINER_BOUNDS_DECLARATIONS `container `elemtype => _BEGIN_
    private
      FHeadItem : P\`container\LinkedItem\`elemtype;
      FTailItem : P\`container\LinkedItem\`elemtype;
      FCount    : integer;
      procedure ResetHeadTail;  begin FHeadItem := nil; FTailItem := nil; FCount := 0; end;
      procedure IncCount;  begin FCount := FCount + 1; end;
      procedure DecCount;  begin FCount := FCount - 1; if FCount = 0 then ResetHeadTail; end;

    protected
      function  InnerNewItem(const value : `elemtype) : P\`container\LinkedItem\`elemtype;
                                begin New(Result); Result.value := value; end;
      function  InnerRemove(FItem : P\`container\LinkedItem\`elemtype) : P\`container\LinkedItem\`elemtype;
          begin
            `next : P\`container\LinkedItem\`elemtype := FItem.Next;
            `prev : P\`container\LinkedItem\`elemtype := FItem.Prev;
            Dispose(FItem);
            if `next <> nil then Result := `next else Result := `prev;
            if `next <> nil then `next.prev := `prev;
            if `prev <> nil then `prev.next := `next;
            DecCount;
          end;
      function  InnerInsertBefore(FItem : P\`container\LinkedItem\`elemtype; const value : `elemtype) : P\`container\LinkedItem\`elemtype;
          begin
            Result  := InnerNewItem(value);
            if FItem = nil then FItem := FHeadItem;
            if FItem = nil
              then begin Result.prev := nil; Result.next := nil; FHeadItem := Result; FTailItem := Result; end
              else begin Result.prev := FItem.Prev; FItem.Prev := Result; Result.Next := FItem;
                         if Result.prev <> nil then Result.prev.Next := Result; end;
            if FHeadItem = FItem then FHeadItem := Result;
            IncCount;
          end;
      function  InnerInsertAfter(FItem : P\`container\LinkedItem\`elemtype; const value : `elemtype) : P\`container\LinkedItem\`elemtype;
          begin
            Result  := InnerNewItem(value);
            if FItem = nil then FItem := FTailItem;
            if FItem = nil
              then begin Result.prev := nil; Result.next := nil; FHeadItem := Result; FTailItem := Result; end
              else begin Result.prev := FItem; Result.Next := FItem.Next; FItem.Next := Result;
                         if Result.Next <> nil then Result.Next.prev := Result; end;
            if FTailItem = FItem then FTailItem := Result;
            IncCount;
          end;
    public
      constructor Create; overload; begin inherited; ResetHeadTail; end;
      destructor Destroy; override; begin clear; inherited; end;

      procedure clear;              begin while not is_empty do FreeItemInList; end;
      function  count    : integer; begin Result := FCount; end;
      function  is_empty : boolean; begin Result := FCount = 0; end;
  _END_;



# DTL_INNERTREESET_ROTATE `container `elem `name `lx `rx => _BEGIN_
      procedure `name(node : P\`container\LinkedItem\`elem);
          begin
            `temp :  P\`container\LinkedItem\`elem := node.`rx;
            node.`rx := `temp.`lx;
            if (`temp.`lx <> nil) then begin `temp.`lx.parent := node; end;
            `temp.parent := node.parent;
            if (node = root)
              then begin froot := `temp; end
              else if (node = node.parent.`lx)
                      then begin node.parent.`lx := `temp; end
                      else begin node.parent.`rx := `temp; end;
            `temp.`lx := node;
            node.parent := `temp;
          end;
  _END_;


# DTL_INNERTREESET_INSERT_FIXUP `px `nx => _BEGIN_
        `y := `zpp.`nx;
        if ((`y <> nil) and (`y.color = Red)) then begin
          `z.parent.color := Black;
          `y.color := Black;
          `zpp.color := Red;
          `z := `zpp;
        end else begin
          if (`z = `z.parent.`nx) then begin `z := `z.parent; rotate\`px(`z); end;
          `z.parent.color := Black;
          `zpp.color := Red;
          rotate\`nx(`zpp);
        end;
  _END_;

# DTL_INNERTREESET_DELETE_FIXUP `container `elemtype `px `nx => _BEGIN_
        `w : P\`container\LinkedItem\`elemtype := `x_parent.`nx;
      if (`w.color = Red)  then begin
        `w.color := Black;
        `x_parent.color := Red;
        rotate\`px(`x_parent);
        `w := `x_parent.`nx;
      end;
      if  (((`w.`px = nil) or (`w.`px.color = Black))
       and ((`w.`nx = nil) or (`w.`nx.color = Black))) then begin
        `w.color := Red;
        `x := `x_parent;
        `x_parent := `x_parent.parent;
      end else begin
        if ((`w.`nx = nil) or (`w.`nx.color = Black)) then begin
          `w.`px.color := Black;
          `w.color := Red;
          rotate\`nx(`w);
          `w := `x_parent.`nx;
        end;
        `w.color := `x_parent.color;
        `x_parent.color := Black;
        if (`w.`nx <> nil)  then begin `w.`nx.color := Black; end;
        rotate\`px(`x_parent);
        `x := root;
      end;
  _END_;

# DTL_INNERTREESET_ITEREATOR_MOVE_NEXT `container `elemtype => _BEGIN_
        if (x.next <> nil) then begin
          x := x.next;
          while (x.prev <> nil) do begin x := x.prev; end;
        end else begin
          `y : P\`container\LinkedItem\`elemtype := x.parent;
          while (x = `y.next) do begin x := `y; `y := `y.parent; end;
          if (x.next <> `y) then x := `y;
        end;
        Result := x;
  _END_;

# DTL_TREE_ITEREATOR_MOVE_PREV `container `elemtype => _BEGIN_
        if (x.prev <> nil)  then begin
          `y : P\`container\LinkedItem\`elemtype := x.prev;
          while (`y.next <> nil) do begin `y := `y.next; end;
          x := `y;
        end else begin
          `y : P\`container\LinkedItem\`elemtype := x.parent;
          while (x = `y.prev) do begin x := `y; `y := `y.parent; end;
          x := `y;
        end;
        Result := x;
  _END_;

# DTL_CONTAINER_TREE_FUNCTIONS `container `elemtype `comp_fn => _BEGIN_
    private
      FRoot : P\`container\LinkedItem\`elemtype;

      function  Minimum(x: P\`container\LinkedItem\`elemtype): P\`container\LinkedItem\`elemtype;
          begin Result := x; while (Result.prev <> nil) do Result := Result.prev; end;
      function  Maximum(x: P\`container\LinkedItem\`elemtype): P\`container\LinkedItem\`elemtype;
          begin Result := x; while (Result.next <> nil) do Result := Result.next; end;
      function  TreeIteratorNext(x : P\`container\LinkedItem\`elemtype) : P\`container\LinkedItem\`elemtype;
          begin DTL_INNERTREESET_ITEREATOR_MOVE_NEXT `container `elemtype; end;
      function  TreeIteratorPrev(x : P\`container\LinkedItem\`elemtype) : P\`container\LinkedItem\`elemtype;
          begin DTL_TREE_ITEREATOR_MOVE_PREV `container `elemtype; end;

      DTL_INNERTREESET_ROTATE `container `elemtype rotatePrev prev next;
      DTL_INNERTREESET_ROTATE `container `elemtype rotateNext next prev;

    private
      function InnerInsertNode(const data : `elemtype) : P\`container\LinkedItem\`elemtype;
          begin
            `y   : P\`container\LinkedItem\`elemtype := nil;
            `x   : P\`container\LinkedItem\`elemtype := root;
            while (`x <> nil) do begin
              `y := `x;
              `cmp : integer := `comp_fn(data, `x.value);
              if (`cmp < 0)
                then begin `x := `x.prev; end
                else if (`cmp > 0)
                        then begin `x := `x.next; end
                        else begin Result := `x; exit; end;
            end;

            IncCount;
            `z : P\`container\LinkedItem\`elemtype := ConsTreeItem(data);
            `z.prev  := nil; `z.next := nil; `z.color := RED;
            Result := `z; 

            if ((FHeadItem = nil) or (`comp_fn(data, FHeadItem.value) < 0))
              then begin FHeadItem := `z; end;
            if ((FTailItem = nil) or (`comp_fn(FTailItem.value, data) < 0))
              then begin FTailItem := `z; end;

            `z.parent := `y;
            if (`y = nil)
              then begin froot := `z; end
              else if (`comp_fn(data, `y.value) < 0)
                      then begin `y.prev := `z; end
                      else begin `y.next := `z; end;

            while ((`z <> root) and (`z.parent.color = Red)) do begin
              `zpp : P\`container\LinkedItem\`elemtype := `z.parent.parent;
              if (`z.parent = `zpp.prev)
                then begin DTL_INNERTREESET_INSERT_FIXUP prev next; end
                else begin DTL_INNERTREESET_INSERT_FIXUP next prev; end;
            end;
            root.color := Black;
          end;

      procedure InnerDeleteNode(z : P\`container\LinkedItem\`elemtype);
          begin
            if z = nil then exit;

            `y : P\`container\LinkedItem\`elemtype := z;
            `x : P\`container\LinkedItem\`elemtype := nil;
            `x_parent : P\`container\LinkedItem\`elemtype := nil;

            if (`y.prev = nil)
              then begin `x := `y.next; end
              else begin if (`y.next = nil)
                            then begin `x := `y.prev; end
                            else begin `y := `y.next; while (`y.prev <> nil) do begin `y := `y.prev; end; `x := `y.next; end;
                   end;

            if (`y <> z) then begin
              z.prev.parent := `y;
              `y.prev := z.prev;
              if (`y <> z.next)
                then begin
                  `x_parent := `y.parent;
                  if (`x <> nil) then begin `x.parent := `y.parent; end;
                  `y.parent.prev := `x;
                  `y.next := z.next;
                  z.next.parent := `y; end
                else begin `x_parent := `y; end;
                
              if (root = z) then begin
                froot := `y;
              end else if (z.parent.prev = z) then begin
                z.parent.prev := `y;
              end else begin
                z.parent.next := `y;
              end;
              `y.parent := z.parent;
              `tmpcol : ENUM_DTL_RB_COLOR := `y.color;
              `y.color := z.color;
              z.color := `tmpcol;
              `y := z;
            end else begin
              `x_parent := `y.parent;
              if (`x <> nil)  then begin `x.parent := `y.parent; end;
              if (froot = z)
                then begin froot := `x; end
                else begin
                  if (z.parent.prev = z)
                    then begin z.parent.prev := `x; end
                    else begin z.parent.next := `x; end;
                end;

              if (FHeadItem = z) then begin
                if (z.next = nil)
                  then begin FHeadItem := z.parent; end
                  else begin FHeadItem := minimum(`x); end;
              end;
              
              if (FTailItem = z) then begin
                if (z.prev = nil)
                  then begin FTailItem := z.parent; end
                  else begin FTailItem := maximum(`x); end;
              end;
            end;

            if (`y.color = Black)  then begin
              while ((`x <> root) and ((`x = nil) or (`x.color = Black))) do begin
                if (`x = `x_parent.prev)
                  then begin DTL_INNERTREESET_DELETE_FIXUP `container `elemtype prev next; end
                  else begin DTL_INNERTREESET_DELETE_FIXUP `container `elemtype next prev; end;
              end;
              if (`x <> nil) then begin `x.color := Black; end;
            end;
            dispose(`y);
            DecCount;
          end;

      function FindNode(const data : `elemtype) : P\`container\LinkedItem\`elemtype;
          begin
            Result := root;
            while (Result <> nil) do begin
              `cmp : integer := `comp_fn(Result.value, data);
              if `cmp < 0
              then begin Result := Result.next; end
              else if `cmp > 0
                      then begin Result := Result.prev; end
                      else begin break; end;
            end;
          end;
    public
      property Root : P\`container\LinkedItem\`elemtype read Result := FRoot init FRoot := nil;

  _END_;


